//------------------------------------------------------------------------------
//
//  npgl.c
//
//  ANTz application based on NPE is free software
//	NPE is the Neural Physics Engine that drives the interactions
//
//	ANTz hosted at http://temporalzone.com and NPE at http://neuralphysics.org
//
//	Copyright (c) 2010-2011, Shane Saxon
//	All rights reserved.
//	Redistribution and use in source and binary forms, with or without 
//	modification, are permitted provided that the following conditions are met:
//		Redistributions of source code must retain the above copyright notice, 
//		this list of conditions and the following disclaimer.
//		Redistributions in binary form must reproduce the above copyright 
//		notice, this list of conditions and the following disclaimer in the 
//		documentation and/or other materials provided with the distribution.
//		Neither the name of the Shane Saxon nor the names of its contributors 
//		may be used to endorse or promote products derived from this software 
//		without specific prior written permission.
//	THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
//	AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
//	IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
//	ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
//	LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
//	CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
//	SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
//	INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
//	CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
//	ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
//	POSSIBILITY OF SUCH DAMAGE.
//
//------------------------------------------------------------------------------

#include "npgl.h"

#include <math.h>

#include "gl/soil/src/SOIL.h"

#include "../npctrl.h"

#include "../os/npos.h"

#include "watchtools.h"
#include "windows.h"									// belongs in os/msw/ debug, zz

unsigned int texture;														// texture, debug zz

int previousX;
int previousY;
int picking;
//MB zz
float theta;
float zeta;
float cmDX;
float cmDY;

#define NOBUTTON -1
int buttonDown;
//MB zz end

void npGLResizeScene (int width, int height);
void npCopyGPUtoCPU (void* dataRef);
void npCameraUpdate( void* dataRef );							//MB zz


void npPick( int x, int y );											//debug, zz //MB//

//------------------------------------------------------------------------------
void npInitGL(void* dataRef) 
{
	previousX = 0;
	previousY = 0;

	picking = false;

	theta = 0.0;
	zeta  = 0.0;
	cmDX = 0.0;
	cmDY = 0.0;

	buttonDown = NOBUTTON;


	npInitGLDraw (dataRef);

	npInitGLDraw (dataRef);

//	glShadeModel (GL_SMOOTH);
//	glClearColor (0.0f, 0.0f, 0.0f, 0.0f);
//	glClearDepth (1.0f);

	glEnable (GL_DEPTH_TEST);

//	glDepthFunc (GL_LEQUAL);
//	glHint (GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);

	  // load our texture												// texture, debug zz
//  texture = LoadTextureRAW( "texture.raw", TRUE );
  texture = SOIL_load_OGL_texture
	(
		"map00001.jpg",
		SOIL_LOAD_AUTO,
		SOIL_CREATE_NEW_ID,
		SOIL_FLAG_MIPMAPS | SOIL_FLAG_INVERT_Y | SOIL_FLAG_NTSC_SAFE_RGB | SOIL_FLAG_COMPRESS_TO_DXT
	);																// texture, debug zz

}


// currently just single window support, add multi-screen capability later... debug zz
// redbook has info on how to calculate the projections
//------------------------------------------------------------------------------
void npGLResizeScene (int wd, int ht)							
{
	int camIndex = 1;	//update with list in data->io->view.camIndex[]		debug zz
	GLfloat ratio = 1.0;

	pData data = npGetDataRef();

	NPnodePtr camNode = data->map.node[camIndex];
	NPcameraPtr camData = camNode->data;

	//MB//						//was width, made global by MB, undo it... debug, zz
	width  = wd;
	height = ht;
	//MB-end//

	if (height > 0)							// Prevent a divide by zero
		camData->aspectRatio = (GLfloat)width / (GLfloat)height;

	// Reset the coordinate system before modifying
	glMatrixMode (GL_PROJECTION);
	glLoadIdentity ();
	
	// Set the viewport to be the entire window
    glViewport(0, 0, width, height);

	// Set the clipping volume
	gluPerspective (camData->fov, camData->aspectRatio, 
					camData->clipNear, camData->clipFar);
	glMatrixMode (GL_MODELVIEW);

	npGLDrawScene (data);
}


//------------------------------------------------------------------------------
void npGLDrawScene (void* dataRef)
{
	int	camIndex = 1;
	GLfloat upX = 0.0f, upY = 0.0f, upZ = 1.0f;
	
	GLfloat position[] = { -150.0f, 150.0f, -150.0f, 1.0f };
	GLfloat positionB[] = { 150.0f, -150.0f, 150.0f, 1.0f };
	GLfloat positionC[] = { 0.0f, 0.0f, 15.0f, 1.0f };

	GLfloat globalAmbient[] = { 0.3f, 0.3f, 0.3f, 1.0f };

	GLfloat materialSpecular[] = { 1.0f, 1.0f, 1.0f, 1.0f };
	GLfloat materialShininess[] = { 128.0f };
	GLfloat materialAmbient[] = { 0.25f, 0.25f, 0.25f, 1.0f };
	GLfloat materialDiffuse[] = { 0.4f, 0.4f, 0.4f, 1.0f };
	GLfloat localAmbient[] = { 0.0f, 0.0f, 0.0f, 1.0f };

	GLfloat angle = 0.0;
	GLfloat redDiffuseMaterial[] = {1.0f, 1.0f, 1.0f}; //set the material to red
	GLfloat whiteSpecularMaterial[] = {1.0f, 1.0f, 1.0f}; //set the material to white
	GLfloat greenEmissiveMaterial[] = {0.1f, 0.1f, 0.1f}; //set the material to green
	GLfloat whiteSpecularLight[] = {1.0f, 1.0f, 1.0f}; //set the light specular to white
	GLfloat blackAmbientLight[] = {0.0f, 0.0f, 0.0f}; //set the light ambient to black
	GLfloat whiteDiffuseLight[] = {1.0f, 1.0f, 1.0f}; //set the diffuse light to white
	GLfloat blankMaterial[] = {0.0f, 0.0f, 0.0f}; //set the diffuse light to white
	GLfloat mShininess[] = {128}; //set the shininess of the material

	GLfloat redNotGreenNotBlue[] = {1.0f, -1.0f, -1.0f}; //set the material to red

	pData data = (pData) dataRef;
	NPnodePtr camNode = data->map.node[camIndex];
	NPcameraPtr camData = camNode->data;

	NPnodePtr gridNode = data->map.node[2];					// texture, debug zz
	NPgridPtr gridData = gridNode->data;

	GLfloat gridX = 0.0f, gridY = 0.0f;

	camNode = camNode->child[camNode->childIndex];

	//lighting
//	glLightModelfv (GL_LIGHT_MODEL_AMBIENT, globalAmbient);
	
	npCtrlUpdate (dataRef);	  //update data, positions, physics, interactions...

	glClearColor (data->io.clear.r, data->io.clear.g, data->io.clear.b,
					data->io.clear.a);
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

//	glfwSwapInterval(3);
	glEnable(GL_DEPTH_TEST);
	glDepthFunc(GL_LESS);

	glEnable (GL_LIGHTING);
	glEnable (GL_LIGHT0);
	glEnable (GL_LIGHT1);
//	glEnable (GL_LIGHT2);		//MB-Transp

	glLightfv (GL_LIGHT0, GL_POSITION, position);
	glLightfv(GL_LIGHT0, GL_SPECULAR, whiteSpecularLight);
    glLightfv(GL_LIGHT0, GL_AMBIENT, blackAmbientLight);
    glLightfv(GL_LIGHT0, GL_DIFFUSE, whiteDiffuseLight);

	glLightfv (GL_LIGHT1, GL_POSITION, positionB);
 
	glLightfv(GL_LIGHT1, GL_SPECULAR, whiteSpecularLight);
    glLightfv(GL_LIGHT1, GL_AMBIENT, blackAmbientLight);
    glLightfv(GL_LIGHT1, GL_DIFFUSE, whiteDiffuseLight);

	glLightfv (GL_LIGHT2, GL_POSITION, positionC);
 
	glLightfv(GL_LIGHT2, GL_SPECULAR, redNotGreenNotBlue);
    glLightfv(GL_LIGHT2, GL_AMBIENT, blackAmbientLight);
    glLightfv(GL_LIGHT2, GL_DIFFUSE, redNotGreenNotBlue);

//	glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, whiteSpecularMaterial);
//	glMaterialfv(GL_FRONT_AND_BACK, GL_SHININESS, mShininess);
	glMaterialfv(GL_FRONT_AND_BACK, GL_DIFFUSE, redDiffuseMaterial);
	glMaterialfv(GL_FRONT_AND_BACK, GL_EMISSION, greenEmissiveMaterial);

	glEnable (GL_COLOR_MATERIAL);

//	glBlendFunc (GL_SRC_COLOR, GL_ONE_MINUS_SRC_ALPHA);
//	glBlendFunc (GL_SRC_ALPHA, GL_DST_COLOR);
	glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);				//MB-Transp
	glEnable (GL_BLEND);													// debug, zz
	//glAlphaFunc (GL_GREATER,0.1f);
//	glEnable (GL_ALPHA_TEST);											//zz
	glEnable (GL_DEPTH_TEST);


	glShadeModel (GL_SMOOTH);


    // setup texture mapping									// texture, debug zz
	gridX = gridNode->scale.x * 5.0f * (GLfloat) gridData->segments.x;
	gridY = gridNode->scale.y * 5.0f * (GLfloat) gridData->segments.y;

	if (!npGetPickPass())	//allow for click on background tex's for mouse flying
	{
	 glDisable (GL_LIGHTING);
      glEnable( GL_TEXTURE_2D );
      glBindTexture( GL_TEXTURE_2D, texture );

	  glColor4f (1.0f, 1.0f, 1.0f, 1.0f);
      glPushMatrix();
      glRotatef( 0.0f, 0.0f, 0.0f, 1.0f );
      glBegin( GL_QUADS );
      glTexCoord2d(0.0,0.0); glVertex3d(-gridX,-gridY,-0.1f);
      glTexCoord2d(1.0,0.0); glVertex3d(+gridX,-gridY,-0.1f);
      glTexCoord2d(1.0,1.0); glVertex3d(+gridX,+gridY,-0.1f);
      glTexCoord2d(0.0,1.0); glVertex3d(-gridX,+gridY,-0.1f);
      glEnd();
      glPopMatrix();
	   glDisable( GL_TEXTURE_2D );
	 glEnable (GL_LIGHTING);										//end texture, debug zz
	}

	npDrawNodes (dataRef);				//draws mesh, graphs, pins, etc...
	
	npCameraUpdate( dataRef );

	glLoadIdentity();

	gluLookAt (	camNode->translate.x, 
				camNode->translate.y,
				camNode->translate.z, 
				camNode->translate.x + camNode->rotate.x,
				camNode->translate.y + camNode->rotate.y,
				camNode->translate.z + camNode->rotate.z,
				upX, upY, upZ );

//	glOrtho ( -2.0f, 2.0f, -2.0f, 2.0f, 0.05f, 2.0f);
//	gluOrtho2D ( -2.0f, 2.0f, -2.0f, 2.0f);

	npDrawHUD (dataRef);				// draws overlays, grids, text...

	npCopyGPUtoCPU (dataRef);			// copies display to main CPU memory

//	glfwSleep (0.1f);
}


// copy the rasterized display to the main system memory (CPU RAM or Drive)
//------------------------------------------------------------------------------
void npCopyGPUtoCPU (void* dataRef)
{
	return;
}


//------------------------------------------------------------------------------
void npMouseEvent (int button, int state, int x, int y)
{
	pData data = (pData) npGetDataRef();

	previousX = x;
	previousY = y;	


	if (button == GLUT_LEFT_BUTTON && state == GLUT_UP)
	{
		data->io.mouseDeltaX = 0.0f;
		data->io.mouseDeltaY = 0.0f;
	}
	

	if( state == GLUT_DOWN )	//Mouse Click Down (Press)
	{
		buttonDown = button;

		PickPass( true );
		npGLDrawScene( data );
		y = height - y;
		npPick( x, y );
		npGLDrawScene( data );							//zz needed to select object... debug
		PickPass( false );
	}
	if( state == GLUT_UP )						//Mouse Click Up (Release)
	{	
		picking = false;
		PickPass( false );
		buttonDown = NOBUTTON;
		cmDY = 0.0;
	}
}

//------------------------------------------------------------------------------
void npMouseMotion (int x, int y)
{
	int dX = 0;
	int dY = 0;
	static float tX = 0.0;
	static float tY = 0.0;
	
	pData data = (pData) npGetDataRef();

	data->io.mouseDeltaX = (GLfloat)(x - previousX);
	data->io.mouseDeltaY = (GLfloat)(y - previousY);


	if( !picking )
	{
		if (data->map.nodeRootCount > 3)
		{
			npCtrlCommand (kGraphSelectAllToggle, data);	//zz debug not efficient
			npCtrlCommand (kGraphCamera, data);
		}

		dX = x - previousX;
		dY = y - previousY;
		if( buttonDown == GLUT_LEFT_BUTTON )
		{
			tX += (float)dX / 300.0f;
			tY += (float)dY / 600.0f;
			theta = tX;
			zeta  = -tY;
		}
		if( buttonDown == GLUT_RIGHT_BUTTON )
		{
			cmDX += (float)dX / 800.0f;
			cmDY += (float)dY / 800.0f;
			
			dX = x - previousX;
			tX += (float)dX / 300.0f;
			theta = tX;
		}
	}

	previousX = x;
	previousY = y;
}

//------------------------------------------------------------------------------
void npPick( int x, int y )
{
	GLubyte r,g,b,a;	//Unsigned Byte 8-bit Range: 0-255 // changed from u_int8_t, zz debug
	char pixels[4];		//4 Bytes
	int idInRGB;

	NPnodePtr node = NULL;
	pData data = (pData) npGetDataRef();;

//	WATCH_FUNC;
//	WATCH_INT( x );

	glReadPixels( x,y, 1,1, GL_RGBA, GL_UNSIGNED_BYTE, pixels );			//Read the Pixel
	
	r = pixels[0];
	g = pixels[1];
	b = pixels[2];
	a = pixels[3];

	idInRGB = RGBtoID( r,g,b );
	selectedID = idInRGB;


	if( selectedID == 0 )
		picking = false;
	else
	{
		picking = true;

		if (selectedID >= kNPnodeMax)	//mouse selection crashing workaround, debug zz
			selectedID = 0;

		node = data->map.nodeID[selectedID];
		if (node != NULL)
		{
			if (node->recordID)
				printf ("recordID: %d \n", node->recordID);
			else
				WATCH_INT( selectedID );
		}
	}
}

//------------------------------------------------------------------------------
void npCameraUpdate( void* dataRef )
{
	int	camIndex = 1;
	pData data = (pData) dataRef;
	NPnodePtr camNode = data->map.node[camIndex];

	camNode = camNode->child[camNode->childIndex];
	
	camNode->rotate.x += 1.0f * (float)sin(theta);
	camNode->rotate.y += 1.0f * (float)cos(theta);
	camNode->rotate.z += zeta;
	
	if( buttonDown == GLUT_RIGHT_BUTTON )
	{
		camNode->translate.x += -cmDY*camNode->rotate.x;
		camNode->translate.y += -cmDY*camNode->rotate.y; 
	}

//	WATCH_FLOAT( camNode->rotate.x );
}

